#!/usr/bin/env python
#-*- coding: ISO-8859-1 -*-

import cherrypy, os, string, time, urllib, traceback
import logging, logging.handlers, popen2

import ConfigParser
from FMWSConfig  import *
from FMWSLogger  import *
from FileManager import Server

# webtools framework
from Framework import Controller
from Framework.PluginManager import DeclarePlugin

from Tools.SecurityModuleCore import encryptCookie, decryptCookie
from Tools.SecurityModuleCore import SecurityDBApi
from Tools.SecurityModuleCore.SecurityDBApi import SecurityDBApi
from Framework import Context
from Framework.Logger import Logger

from Tools.SecurityModuleCore import SecurityToken, RedirectToLocalPage, RedirectAway, RedirectorToLogin
from Tools.SecurityModuleCore import Group, Role, NotAuthenticated, FetchFromArgs
from Tools.SecurityModuleCore import is_authorized, is_authenticated, has_site
from Tools.Functors import AlwaysFalse


def makeConfigParser():
    area = ""
    try:
        area = os.environ['FMWSTRANSFERDIR']
    except:
        raise "FMWSTRANSFERDIR environment is not set up"
    cp = ConfigParser.ConfigParser()
    cp.add_section("file_manager")
    cp.set("file_manager", "base_directory", "%s"%area)
#    cp.add_section("max_size_gb")
    cp.set("file_manager","max_size_gb", "20")
    cp.add_section("transfer_wrapper")
    cp.set("transfer_wrapper", "transfer_command", "srmcp -debug=true -srm_protocol_version=2 -retry_num=1 -streams_num=1")
    cp.add_section("file_lookup")
    cp.set("file_lookup","priority_0","T1_US")
    cp.set("file_lookup","priority_1","T2")
    cp.set("file_lookup","priority_2","T1")
    return cp

class FMWS(FMWSLogger,Controller):
    def __init__(self,context=""):
        self.securityApi=""
        try:
            if context:
               context.OptionParser ().add_option("-v","--verbose",action="store",type="int", 
               default=0, dest="verbose",help="specify verbosity level, 0-none, 1-info, 2-debug")
               self.securityApi=SecurityDBApi(context)
            Controller.__init__ (self, context, __file__)
            self.config=self.setConfig(base="FMWS")
        except:
#            traceback.print_exc()
            pass
        cp = makeConfigParser()
        Server.configure(cp)
        self.fmConfig = FMWSConfig()
        self.verbose= self.fmConfig.verboseLevel()
#        self.voms_timer=time.time()
        self.voms_timer=0
        self.userDict={}
        FMWSLogger.__init__(self,self.fmConfig.loggerDir(),"FMWSServer",self.verbose)
        setLogger('cherrypy.access',super(FMWS,self).getHandler(),super(FMWS,self).getLogLevel())
        setLogger('cherrypy.error',super(FMWS,self).getHandler(),super(FMWS,self).getLogLevel())
        setLogger('ThreadPool',super(FMWS,self).getHandler(),super(FMWS,self).getLogLevel())
        setLogger('TransferWrapper',super(FMWS,self).getHandler(),super(FMWS,self).getLogLevel())
        setLogger('FileLookup',super(FMWS,self).getHandler(),super(FMWS,self).getLogLevel())

    def readyToRun(self):
        opts=self.context.CmdLineArgs().opts
#        self.mastheadUrl=self.baseUrl+"sitedb/Common/masthead"
#        self.footerUrl=self.baseUrl+"sitedb/Common/footer"
#        self.topUrl=self.baseUrl
        try:
           self.verbose=opts.verbose
        except:
           pass

    def index(self,user="USER"):
        page = self.getTopHTML()
#        if not self.checkUser(user):
        userName = self.getUserName()
#        print "\nuserName",userName
        if not userName or userName=="guest" or userName.lower()=="unknown":
           page+="<p>In order to use CMS File service you must login with your HyperNews login name and password.</p><p>Please follow Login link in right upper corner of the page.</p>"
        else:
           user=userName
           self.addUser(user)
           page+= self.userForm(user)
        page+= self.getBottomHTML()
        return page
    index.exposed=True

    def renewCert(self):
        if (time.time()-self.voms_timer)>12*60*60: # more then 12 hours
           self.voms_timer=time.time()
#           res = popen2.Popen4("voms-proxy-init -voms cms -q")
#           res = popen2.Popen4("cat $HOME/.globus/pp.txt | grid-proxy-init -pwstdin -q")
           cmd="cat $FMWSHOME/pp.txt | grid-proxy-init -cert /data/vk/cert/usercert.pem -key /data/vk/cert/userkey.pem -pwstdin -q"
           res = popen2.Popen4(cmd)
           res.wait()
           result=res.fromchild.read()
           if result:
              msg="Fail to initialize voms, '%s'"%result
              raise msg

    def addUser(self,user):
        if not self.userDict.has_key(user):
           self.userDict[user]=([],[])
           try:
               os.makedirs("%s/download/%s/softlinks"%(os.getcwd(),user))
           except:
               pass

    def checkUser(self,user):
        # TODO: check user against SiteDB.
        if user=="USER" or user=="vk":
           return 1 # valid user
        return 0    # invalid user
        
    def getTopHTML(self):
        page = """
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" lang="en" xml:lang="en">
<head profile="http://www.w3.org/2005/11/profile">
<title>CMS File Server</title>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
<meta http-equiv="Content-Language" content="en-us" />
<meta name="author" content="Valentin Kuznetsov (vkuznet at gmail dot com) and Brian Bockelman" />
<meta name="MSSmartTagsPreventParsing" content="true" />
<meta name="ROBOTS" content="ALL" />
<meta name="Copyright" content="(CC) 2008, CMS collaboration." />
<meta http-equiv="imagetoolbar" content="no" />
<meta name="Rating" content="General" />
<script type="text/javascript" src="js/prototype.js"></script>
<script type="text/javascript" src="js/rico.js"></script>
<script type="text/javascript" src="js/util.js"></script>


<link rel="stylesheet" type="text/css" href="https://cmsweb.cern.ch/sitedb/Common/mastheadcss?site=phedex"/>

<script type="text/javascript" src="https://cmsweb.cern.ch/sitedb/YUI/yui/js/?script=yahoo.js&script=dom.js&script=event.js&script=container_core.js&script=connection.js&script=dragdrop.js&script=container.js"></script>
<script type="text/javascript" src="https://cmsweb.cern.ch/sitedb/Common/masthead"></script>

<style type="text/css">
table.fmws_table {
  width: 100%;
  font-size: 12px;
  font-weight: normal;
  font-family: verdana, sans-serif;
  text-decoration: none;
  border-width: 1px 1px 1px 1px;
  border-spacing: 3px;
  border-style: solid solid solid solid;
  border-color: gray gray gray gray;
  border-collapse: collapse;
  background-color: rgb(255, 250, 250);
}
table.fmws_table th {
  text-transform: uppercase;
  font-size: 1.1em;
  font-weight: bold;
  padding: 5px;
  margin: 0px;
  background-color: #F0F0F0;
  vertical-align: top;
  text-align: right;
}

table.fmws_table th.th_left {
  text-align: left;
}
table.fmws_table th.middle {
  text-align: middle;
}

table.fmws_table tr.tr_th {
  background-color: #F0F0F0;
  border-width: 1px 1px 1px 1px;
  border-style: solid solid solid solid;
  border-color: gray gray gray gray;
  padding: 5px;
  margin: 0px;
}
table.fmws_table tr {
  background-color: #FFFADC;
  border-width: 1px 1px 1px 1px;
  border-style: solid solid solid solid;
  border-color: gray gray gray gray;
  padding: 5px;
  margin: 0px;
}
table.fmws_table tr.zebra{
  background-color: #fff5e8;
  border-width: 1px 1px 1px 1px;
  border-style: solid solid solid solid;
  border-color: gray gray gray gray;
  padding: 5px;
  margin: 0px;
}

table.fmws_table td {
  text-align: right;
  padding: 5px;
  margin: 0px;
}

table.fmws_table td.td_left {
  text-align: left;
  padding: 5px;
  margin: 0px;
}
</style>
<!--
<body>
-->

<body onload="loadMasthead();" id="content">
        """
        return page

    def getBottomHTML(self):
        page = """
</body></html>
        """
        return page

    def formatLFNList(self,user,iList,sList):
        page=""
        style=""
        for idx in xrange(0,len(iList)):
            lfn = iList[idx]
            stat= sList[idx]
            if style: style=""
            else:     style="class=\"zebra\""
            spanid = self.spanId(lfn)
            page+="""<tr %s><td class="td_left">
%s
</td><td align="right">
<span id="%s">%s</span>
</td></tr>
<script type="text/javascript">
var lfnUpdater = new Updater('%s');
ajaxEngine.registerAjaxObject('%s',lfnUpdater);
</script>
"""%(style,lfn,spanid,stat,spanid,spanid)
#            if stat:
#               page+="<img src=\"images/loading.gif\" /> "
            if stat:
               page+="""<script type="text/javascript">setTimeout('ajaxStatusOne(\\'%s\\',\\'%s\\')',1000)</script>"""%(user,lfn)
        return page
        
    def getStat(self,user):
        try:
           iList,sList = self.userDict[user]
           iList.reverse() # reverse list to see newly requested files first
           sList.reverse()
        except:
           pass
           return ""
        page = """
<br />
<div>
<table class="fmws_table">
%s
</table>
</div>
        """%self.formatLFNList(user,iList,sList)
        return page
        
    def getStatForLfn(self,user,lfn):
        page=""
        style=""
        lfnList,statList = self.userDict[user]
        try:
            idx = lfnList.index(lfn)
            stat = statList[idx]
            spanid = self.spanId(lfn)
            page+="""<span id="%s">%s</span>"""%(spanid,stat)
            if stat.find("Download")==-1:
               page+=" | <a href=\"javascript:ajaxCancel('%s','%s')\">Cancel</a> "%(user,lfn)
        except:
            traceback.print_exc()
            print lfn
            print self.userDict
            pass
        return page
        
    def userForm(self,user):
        page = """
<hr style="border:none;border-top: 1px dotted #CCCCCC;" />
Hello %s,
<hr style="border:none;border-top: 1px dotted #CCCCCC;" />
<p />
<b>Request LFN transfer</b>
<form action="javascript:ajaxAction('%s')" method="get">
<input id="lfn" name="lfn" type="text" size="95" />
<input id="method" type="hidden" />
<p />
<input type="submit" value="Request" onclick="javascript:SetMethod('request')" />
<!--
<input type="submit" value="Status"  onclick="javascript:SetMethod('status')" />
<input type="submit" value="Cancel"  onclick="javascript:SetMethod('cancel')" />
-->
<input type="reset" value="Reset" />
</form>
<hr style="border:none;border-top: 1px dotted #CCCCCC;" />
<div id="_response">%s</div>
        """%(user,user,self.checkUserCache(user) )
        return page

    def ajaxResponse(self,msg,tag="_response"):
        page ="""<ajax-response><response type="object" id="%s">"""%tag
        page+=msg
        page+="</response></ajax-response>"
        return page

    def spanId(self,lfn):
        return lfn.split("/")[-1].replace(".root","")
        
    def nActiveDownloads(self,user):
        lfnList,statList = self.userDict[user]
        count=0
        for stat in statList:
            if stat.find("Download")==-1:
               count+=1
        return count
        
    def addLfn(self,user,lfn):
        lfnList,statList = self.userDict[user]
        if len(lfnList)<int(self.fmConfig.maxTransfer()):
#        if self.nActiveDownloads(user)<int(self.fmConfig.maxTransfer()):
           if  not lfnList.count(lfn):
               lfnList.append(lfn)
               statList.append('requested')
           else:
              return 2
        else: 
           return 0
        self.userDict[user]=(lfnList,statList)
        return 1
        
    def delLfn(self,user,lfn):
        lfnList,statList = self.userDict[user]
        if lfnList.count(lfn):
           idx = lfnList.index(lfn)
           lfnList.remove(lfn)
           statList.remove(statList[idx])
           self._remove(user,lfn)

    def checkUserCache(self,user):
        page = ""
        lfnList,statList = self.userDict[user]
        pfnList = os.listdir("%s/download/%s/softlinks"%(os.getcwd(),user))
        for file in pfnList:
            f = "%s/download/%s/softlinks/%s"%(os.getcwd(),user,file)
            abspath= os.readlink(f)
            fileStat = os.stat(abspath)
            fileSize = self.sizeFormat(fileStat[stat.ST_SIZE])
            link     = "download/%s/%s"%(user,file)
            lfn      = abspath.replace(self.fmConfig.transferDir(),"")
            msg="<a href=\"https://cmsweb.cern.ch/filemover/%s\">Download (%s)</a> | <a href=\"javascript:ajaxRemove('%s','%s')\">Remove</a> "%(link,fileSize,user,lfn)
#            msg="<a href=\"%s\">Download (%s)</a>"%(link,fileSize)
            if  not lfnList.count(lfn):
                lfnList.append("%s"%lfn)
                statList.append(msg)
        self.userDict[user]=(lfnList,statList)
        page+= self.getStat(user)
        return page
        
    def sizeFormat(self,i):
        """
           Format file size utility, it converts file size into KB, MB, GB, TB, PB units
        """
        try:
           num=long(i)
        except:
           return "N/A"
        for x in ['','KB','MB','GB','TB','PB']:
            if num<1024.:
                return "%3.1f%s" % (num, x)
            num /=1024.

    
    def setStat(self,user,_lfn=None):
        statCode =-1
        if not self.userDict.has_key(user):
           self.userDict[user]=([],[])
        lfnList,statList = self.userDict[user]
        for idx in xrange(0,len(lfnList)):
            lfn    = lfnList[idx]
            if _lfn and lfn!=_lfn: continue
            prevStat = statList[idx]
#            status = (0,'complete')
            status = Server.status(lfn)
#            print "\nstatus",lfn,status
            if  status=="This LFN has not been requested yet!":
                status = (0,'orphan')
            if  lfn.find("/")==-1:
                statCode = 0
                statMsg  = prevStat 
                statList[idx]=statMsg
                file = lfn
                pfn= "%s/download/%s/%s"%(os.getcwd(),user,file)
            else:
                statCode = status[0]
                statMsg  = status[1]
                statList[idx]=statMsg
                iList  = lfn.split("/")
                file   = iList[-1]
                dir    = self.fmConfig.transferDir()+'/'.join(iList[:-1])
                pfn    = os.path.join(dir,file)
            if statCode==0:
               if  os.path.isfile(pfn):
                   if not os.path.isfile("%s/download/%s/%s"%(os.getcwd(),user,file)):
                      os.link(pfn,"%s/download/%s/%s"%(os.getcwd(),user,file))
                      os.symlink(pfn,"%s/download/%s/softlinks/%s"%(os.getcwd(),user,file))
                   link     = "download/%s/%s"%(user,file)
#                   rlink    = "remove?file=%s"%file 
                   fileStat = os.stat(link)
                   fileSize = self.sizeFormat(fileStat[stat.ST_SIZE])
                   lfn      = pfn.replace(self.fmConfig.transferDir(),"")
                   msg="<a href=\"https://cmsweb.cern.ch/filemover/%s\">Download (%s)</a> | <a href=\"javascript:ajaxRemove('%s','%s')\">Remove</a> "%(link,fileSize,user,lfn)
#                   msg="<a href=\"%s\">Download (%s)</a>"%(link,fileSize)
                   statList[idx]=msg
               else:
                   statList[idx]="Download error, <a href=\"javascript:ajaxRemove('%s','%s')\">Remove</a> "%(user,lfn)
        self.userDict[user]=(lfnList,statList)
        return statCode

    def getArg(self,kwargs,key,default):
        arg=default
        if kwargs.has_key(key):
           try:
              arg=kwargs[key]
              if type(default) is types.IntType:
                 arg=int(arg)
           except:
              pass
        return arg

    def getUserName(self):
        """
           Get userName from stored cookie, should be run within WEBTOOLS framework
        """
        userName=""
        try:
            cookie=cherrypy.request.cookie
            userName = decryptCookie (cookie["dn"].value, self.securityApi)
        except:
            traceback.print_exc()
            pass
        return userName

    def _remove(self,user,lfn):
        file = lfn.split("/")[-1]
        # remove soft-link from user download area
        try:
            link = "%s/download/%s/softlinks/%s"%(os.getcwd(),user,file)
            os.unlink(link)
        except:
#            traceback.print_exc()
            pass
        # remove hard-link from user download area
        try:
            link = "%s/download/%s/%s"%(os.getcwd(),user,file)
            os.unlink(link)
        except:
#            traceback.print_exc()
            pass
        # now time to check if no-one else has a hardlink to pfn, if so remove pfn
        try:
            pfn = self.fmConfig.transferDir()+lfn
            fstat = os.stat(pfn)
#            print "fstat[stat.ST_NLINK]",pfn,fstat[stat.ST_NLINK]
            if int(fstat[stat.ST_NLINK])==1: # only 1 file exists and no other hard-links to it
               os.remove(pfn)
        except:
#            traceback.print_exc()
            pass

    def remove(self,user,lfn,**kwargs):
        cherrypy.response.headers['Content-Type']='text/xml'
        page = ""
        self._remove(user,lfn)
        self.delLfn(user,lfn)
        page+=self.getStat(user)
        return self.ajaxResponse(page)
    remove.exposed=True

    def request(self,user,lfn,**kwargs):
        cherrypy.response.headers['Content-Type']='text/xml'
        self.renewCert()
#        print "\n### request",user,lfn,kwargs
        lfn = lfn.strip() # remove spaces around lfn
        lfnStatus = self.addLfn(user,lfn)
        if not lfnStatus:
           page ="<p>Too many requests, you're allowed to fetch only %s LFNs</p>"%self.fmConfig.maxTransfer()
           page+= self.getStat(user)
           return self.ajaxResponse(page)
        lfn = urllib.unquote(lfn)
        page=""
        try:
            if lfnStatus==1:
               Server.request(lfn)
               page="<div>Requested file:<p><em>%s</em></p> has been placed into the transfer queue</div>"%lfn
            else:
               page="<div>Requested file:<p><em>%s</em></p> is already in a transfer queue</div>"%lfn
            page+=self.getStat(user)
        except:
            msg = traceback.format_exc()
            self.writeLog(msg)
            page="<div>Request failed</div>"
            page+="<pre>%s</pre>"%msg
        page = self.ajaxResponse(page)
#        print page
        return page
    request.exposed=True
        
    def cancel(self,user,lfn,**kwargs):
        cherrypy.response.headers['Content-Type']='text/xml'
        self.delLfn(user,lfn)
        lfn = urllib.unquote(lfn)
        page=""
        try:
            Server.cancel(lfn)
            page ="<div>Requested file:<p><em>%s</em></p> has been discarded from the transfer queue</div>"%lfn
        except:
            self.writeLog(traceback.format_exc())
            page ="<div>Request failed</div>"
        page+=self.getStat(user)
        page = self.ajaxResponse(page)
        return page
    cancel.exposed=True

#    def status(self,lfn,**kwargs):
#        cherrypy.response.headers['Content-Type']='text/xml'
#        user= self.getArg(kwargs,'user','USER')
#        lfn = urllib.unquote(lfn)
#        page=""
#        try:
#            self.setStat(user)
#            page+=self.getStat(user)
#        except:
#            self.writeLog(traceback.format_exc())
#            page="<div>Request failed</div>"
#        page = self.ajaxResponse(page)
#        return page
#    status.exposed=True

    def statusOne(self,user,lfn,**kwargs):
#        print "\n\n### statusOne",user,lfn
        cherrypy.response.headers['Content-Type']='text/xml'
        lfn = urllib.unquote(lfn)
        lfn = lfn.strip()
        page=""
        try:
            statCode = self.setStat(user,lfn)
            if statCode:
               page+="<img src=\"images/loading.gif\" /> "
            page+=self.getStatForLfn(user,lfn)
            if statCode:
               page+="""<script type="text/javascript">setTimeout('ajaxStatusOne(\\'%s\\',\\'%s\\')',1000)</script>"""%(user,lfn)
        except:
            msg=traceback.format_exc()
            self.writeLog(msg)
            page="<div>Request failed</div>"
        page = self.ajaxResponse(page,self.spanId(lfn))
#        print "\n+++statusOne",page
        return page
    statusOne.exposed=True

    def setConfig(self,base=""):
        mime_types=['text/css','text/javascript','application/javascript','application/x-javascript','image/gif','image/png','image/jpg','image/jpeg','application/octet-stream']
        httpHeader=[('Expires',time.strftime("%a, %d %b %Y %H:%M:%S GMT",time.gmtime(time.time()+315360000))),
                                    ('Accept-Encoding','gzip'),
                                    ('Cache-Control','max-age=315360000'),
                                    ('Authorization','Basic')
                        ]

        conf={'/': {'tools.staticdir.root': os.getcwd(),
                    'tools.response_headers.on':True,
                    'tools.response_headers.headers':
                      [('Expires','Thu, 15 Apr 2010 20:00:00 GMT'),
                       ('Cache-Control','no-store, no-cache, must-revalidate,post-check=0, pre-check=0')],
                   },
              '/js'       : {'tools.gzip.on': True, 
                             'tools.gzip.mime_types':mime_types,
                             'tools.staticdir.on':True,
                             'tools.staticdir.dir':'js',
                             'tools.staticdir.content_types':{'js':'text/javascript'},
                             'tools.response_headers.on':True,
                             'tools.response_headers.headers':httpHeader
                            },
              '/css'      : {'tools.gzip.on': True, 
                             'tools.gzip.mime_types':mime_types,
                             'tools.staticdir.on':True,
                             'tools.staticdir.root': os.getcwd(),
                             'tools.staticdir.dir':'css',
                             'tools.response_headers.on':True,
                             'tools.response_headers.headers':httpHeader
                            },
              '/download' : {
                             'tools.staticdir.on':True,
                             'tools.staticdir.root': os.getcwd(),
                             'tools.staticdir.dir':'download',
                             'tools.staticdir.content_types':{'root':'application/octet-stream'},
                             'tools.response_headers.on':True,
                             'tools.response_headers.headers':httpHeader
                            },
              '/images'   : {'tools.gzip.on': True, 
                             'tools.gzip.mime_types':mime_types,
                             'tools.staticdir.on':True,
                             'tools.staticdir.root': os.getcwd(),
                             'tools.staticdir.dir':'images',
                             'tools.response_headers.on':True,
                             'tools.response_headers.headers':httpHeader
                            },

             }
        if  base:
            newConf={}
            for key in conf.keys():
                newConf[key]=conf[key]
                if key=="/":
                   newKey="/%s"%base.replace("/","")
                else:
                   newKey="/%s/%s"%(base.replace("/",""),key.replace("/",""))
                newConf[newKey]=conf[key]
            return newConf
        return conf


if __name__ == '__main__':

   cherrypy.response.headers['Expires']="Thu, 15 Apr 2010 20:00:00 GMT"
   cherrypy.response.headers['Cache-Control']="no-store, no-cache, must-revalidate, max-age=315360000"
   cherrypy.response.headers['Cache-Control']="post-check=0, pre-check=0"
   port = 8201
   cherrypy.config.update({'server.socket_port': port,
                           'server.thread_pool': 20,
                           'environment': 'production',
                           'log.screen':True,
                           'log.error_file':"fmws.log",
                          })
   fmws=FMWS()
   cherrypy.quickstart(fmws, '/', config=fmws.setConfig())
